from keras.callbacks import Callback
import json
from collections import OrderedDict


class Callback_Image_Classification(Callback):
    def __init__(self, total_step, total_epoch):
        self.step = 0
        self.epoch = 0
        self.total_step = total_step
        self.total_epoch = total_epoch

    def on_batch_end(self, batch, logs={}):
        self.step += 1
        logJson = OrderedDict()
        logJson['logType'] = "batchend"
        logJson['totalStep'] = self.total_step
        logJson['nowStep'] = self.step
        logJson['totalEpoch'] = self.total_epoch
        logJson['nowEpoch'] = self.epoch
        logJson['loss'] = str(logs.get('loss'))
        logJson['acc'] = str(logs.get('acc'))
        print(json.dumps(logJson, ensure_ascii=False, indent="\t"))
        return
    
    def on_epoch_end(self, epoch, logs={}):
        self.epoch += 1
        self.step = 0
        logJson = OrderedDict()
        logJson['logType'] = "epochend"
        logJson['totalEpoch'] = self.total_epoch
        logJson['nowEpoch'] = self.epoch
        logJson['loss'] = str(logs.get('loss'))
        logJson['acc'] = str(logs.get('acc'))
        print(json.dumps(logJson, ensure_ascii=False, indent="\t"))
        return


from __future__ import division
from __future__ import print_function

import keras

from keras.applications import resnet50, densenet, inception_v3, inception_resnet_v2, mobilenetv2, nasnet, xception
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import multi_gpu_model
import numpy as np
import cv2
from math import ceil
import os
import json
from collections import OrderedDict

from ....lib.general_lib import reverse_dict, DataGenerator
from ....lib.image_lib import directory_images_to_arrays
from ....lib.image_proccessing.image_classification.saving_lib import load_model, save_model
from ....lib.custom_callback import Callback_Image_Classification

from ...application import Application



class Image_Classification(Application):
    def __init__(self, model_path = None, model_type = 'densenet121'):
        super().__init__()
        if type(model_path) != type(None):
            self.load(model_path)
        else:
            self.model_type = model_type
    

    def prepare_train_data(self, get_image_from = 'directory', data_path = None, data_array = None, target_array = None, optimizer = 'nadam', metrics = ['accuracy']):
        self.get_image_from = get_image_from
        self.data_path = data_path
        if hasattr(self, 'model'):
            if self.get_image_from == 'argument':
                return self.reshape_data(data_array), target_array
            else:
                return
        
        if self.get_image_from == 'argument':
            if type(data_array) == type(None) or type(target_array) == type(None):
                raise TypeError('you should prepare arrays to initialize model')
            else:
                unique_classes = np.lib.arraysetops.unique(target_array)
                class_indices = self.create_class_indices(unique_classes)
                self.create_and_compile_model(class_indices, optimizer, metrics)
                return self.reshape_data(data_array), target_array
        else:
            if self.get_image_from != 'directory':
                raise ValueError("value of get_image_from should be either 'directory' or 'argument'.")
            else:
                class_indices = {}
                for i, c in enumerate(sorted(os.listdir(data_path))):
                    if os.path.isdir(data_path + '/' + c):
                        class_indices[c] = i
                self.create_and_compile_model(class_indices, optimizer, metrics)
                return


    def compile_model(self, optimizer, loss, metrics=['accuracy']):
        self.model.compile(optimizer, loss=loss, metrics=metrics)


    def create_and_compile_model(self, class_indices = None, optimizer = 'nadam', metrics = ['accuracy']):
        num_classes = len(class_indices)
        if self.model_type == 'densenet121':
            self.model = densenet.DenseNet121( weights=None, classes=num_classes )
        elif self.model_type == 'inception_v3':
            self.model = inception_v3.InceptionV3( weights=None, classes=num_classes )
        elif self.model_type == 'inception_resnet_v2':
            self.model = inception_v3.InceptionV3( weights=None, classes=num_classes )
        elif self.model_type == 'mobilenet_v2':
            self.model = mobilenetv2.MobileNetV2( weights=None, classes=num_classes )
        elif self.model_type == 'nasnet_large':
            self.model = nasnet.NASNetLarge( weights=None, classes=num_classes )
        elif self.model_type == 'nasnet_mobile':
            self.model = nasnet.NASNetMobile( weights=None, classes=num_classes )
        elif self.model_type == 'xception':
            self.model = xception.Xception( weights=None, classes=num_classes )
        elif self.model_type == 'resnet50' :
            self.model = resnet50.ResNet50( weights=None, classes=num_classes)
        try:
            self.model = multi_gpu_model(self.model, gpus = self.gpu_num, cpu_relocation=True)
        except ValueError:
            self.model = self.model        
        self.model.class_indices = class_indices
        self.compile_model(optimizer = optimizer, loss = keras.losses.categorical_crossentropy, metrics = metrics)


    def train(self,
            x_train = None,
            target_array = None,
            image_data_generator = None,
            batch_size = 32,
            epochs = 1,
            verbose = 1,
            callbacks = None,
            shuffle = True
        ):

        if type(image_data_generator) == type(None):
            if self.get_image_from == 'directory':
                train_datagen = self.create_image_data_generator(rescale = 1./255)
            else:
                train_datagen = self.create_image_data_generator()
        else:
            train_datagen = image_data_generator
        if self.get_image_from == 'directory':
            train_generator = train_datagen.flow_from_directory(self.data_path, target_size = self.model.input_shape[1:3], class_mode = 'categorical', batch_size = batch_size)
            num_dataset = 0
            for root, dirs, files in os.walk(self.data_path):
                for basename in files:
                    file_extension = basename.split('.')[-1]
                    if file_extension in ('jpg', 'jpeg', 'png', 'bmp', 'ppm', 'tif', 'tiff'):
                        num_dataset += 1
            steps_per_epoch = ceil(num_dataset / batch_size)
            callbacks = [Callback_Image_Classification(total_step = steps_per_epoch, total_epoch = epochs)]
            self.model.fit_generator(train_generator, steps_per_epoch=steps_per_epoch, epochs=epochs, verbose=verbose, callbacks=callbacks, shuffle=shuffle)
        elif self.get_image_from == 'argument':
            y_train = self.prepare_y( target_array, reverse_dict(self.model.class_indices), len(self.model.class_indices) )
            train_generator = train_datagen.flow(x_train, y_train, batch_size = batch_size)
            num_dataset = x_train.shape[0]
            steps_per_epoch = ceil(num_dataset / batch_size)
            self.model.fit_generator(train_generator, steps_per_epoch=steps_per_epoch, epochs=epochs, verbose=verbose, callbacks=callbacks, shuffle=shuffle)


    def load(self, model_path = None):
        try:
            self.model = load_model(model_path)
        except Exception as e:
            print("Model loading failed")
            print(e)
            raise e


    def save(self, model_path):
        try:
            if type(self.model) == type(None):
                raise TypeError('You should create a model before saving it')
            save_model(self.model, model_path, class_indices = self.model.class_indices)
        except Exception as e:
            raise e


    def predict(self, data_array = None, data_path = None, predict_classes = True, batch_size = 16, verbose = 0, steps = None):
        try:
            if type(data_path) == type(None):
                raise TypeError('You sholud give a proper data path')

            output_dictionary = reverse_dict(self.model.class_indices)

            pair_dict = OrderedDict()
            # check data_path is file or folder
            if os.path.isdir(data_path):
                datagen = DataGenerator()
                
                predict_data_generator = datagen.flow_one_directory(
                    directory_path = data_path,
                    shuffle = False,
                    batch_size = batch_size,
                    image_shape = self.model.input_shape[1:3]
                )

                nb_files = len(predict_data_generator.image_filenames)
                probs = self.model.predict_generator(predict_data_generator, steps=ceil(nb_files/batch_size))
                predicted_classes = probs.argmax(axis=-1)

                predicted_order_filenames = predict_data_generator.predict_order_filenames

                for i in range(len(predicted_classes)):
                    pair_dict[predicted_order_filenames[i]] = output_dictionary[predicted_classes[i]]
                return json.dumps(pair_dict, ensure_ascii=False, indent="\t")
            else:
                _, only_file_name = os.path.split(data_path)
                img_to_predict = cv2.imread(data_path)
                img_array = np.expand_dims(img_to_predict, axis=0)
                reshaped_array = self.reshape_data(img_array)
                probs = self.model.predict(reshaped_array, batch_size=batch_size, verbose=verbose, steps=steps)
                predicted_class = probs.argmax(axis=-1)
                pair_dict[only_file_name] = output_dictionary[predict_classes]
                return json.dumps(pair_dict, ensure_ascii=False, indent="\t")
        except Exception as e:
            raise e

    
    # You have to give data_path exactly same as train dataset
    def evaluate(self, data_path = None):
        try:
            total_nb = 0
            total_right_predict = 0
            class_folders = os.listdir(data_path)
            for i in range(len(class_folders)):
                right_predict = 0
                one_class_folder_path = data_path + '/' + class_folders[i]  
                answer = class_folders[i]
                predict_list = self.predict(data_path = one_class_folder_path)
                for j in range(len(predict_list)):
                    if predict_list[j] == answer:
                        right_predict = right_predict + 1
                        total_right_predict = total_right_predict + 1

                print(answer + ":" + str(right_predict/len(predict_list)))

                total_nb = total_nb + len(predict_list)
            print("Total predict evaluation result : " + str(total_right_predict/total_nb))
        except Exception as e:
            raise e


    def check_one_image_in_folder(self, data_path = None):
        try:
            if not os.path.isdir(data_path):
                raise FileNotFoundError(data_path + " does not exist")
            files = os.listdir(data_path)
            evaulate_file = files[0]
            predicted_data = self.predict( data_path = os.path.join(data_path, evaulate_file) )
            return predicted_data.item(0)
        except Exception as e :
            raise e


    def create_class_indices(self, unique_classes):
        class_indices = dict()
        for i in range(unique_classes.shape[0]):
            class_indices[unique_classes[i]] = i
        return class_indices


    def create_image_data_generator(self, featurewise_center=False,  # set input mean to 0 over the dataset
            samplewise_center=False,  # set each sample mean to 0
            featurewise_std_normalization=False,  # divide inputs by std of the dataset
            samplewise_std_normalization=False,  # divide each input by its std
            zca_whitening=False,  # apply ZCA whitening
            zca_epsilon=1e-06,  # epsilon for ZCA whitening            
            rotation_range=0.,  # randomly rotate images in the range (degrees, 0 to 180)
            width_shift_range=0.,  # randomly shift images horizontally (fraction of total width)
            height_shift_range=0.,  # randomly shift images vertically (fraction of total height)
            brightness_range=None,
            shear_range=0.,  # set range for random shear(Shear angle in counter-clockwise direction in degrees)
            zoom_range=0.,  # set range for random zoom. If a float, `[lower, upper] = [1-zoom_range, 1+zoom_range]`.
            channel_shift_range=0.,  # set range for random channel shifts
            fill_mode='nearest',  # One of {"constant", "nearest", "reflect" or "wrap"}
            cval=0.,  # value used for fill_mode = "constant"
            horizontal_flip=True,  # randomly flip images
            vertical_flip=False,  # randomly flip images
            rescale=None,  # set rescaling factor (applied before any other transformation)
            preprocessing_function=None,  # set function that will be applied on each input
            data_format="channels_last",  # either "channels_first" or "channels_last"
            validation_split=0.0  # fraction of images reserved for validation (strictly between 0 and 1)
        ):
            return ImageDataGenerator(featurewise_center,
                samplewise_center,
                featurewise_std_normalization,
                samplewise_std_normalization,
                zca_whitening,
                zca_epsilon,
                rotation_range,
                width_shift_range,
                height_shift_range,
                brightness_range,
                shear_range,
                zoom_range,
                channel_shift_range,
                fill_mode,
                cval,
                horizontal_flip,
                vertical_flip,
                rescale,
                preprocessing_function,
                data_format,
                validation_split)


    def prepare_y(self, target_array, class_indices, num_classes):
        """Prepares y for a keras model.(ex. y_train)
        
        Arguments:
            target_array {ndarray} -- ndarray. Appropriate outputs of input data.
            class_indices {dict} -- A dictionary of classes to targets.
            num_classes {int} -- A number of the entire classes.
        
        Returns:
            [ndarray] -- A binary matrix representation of the input.
        """
        y_list = list()
        for i in range(target_array.shape[0]):
            label_class = class_indices[target_array[i]]
            y_list.append(label_class)        
        y = keras.utils.to_categorical(np.asarray(y_list), num_classes)
        return y


    def reshape_data(self, data_array, cv_interpolation = cv2.INTER_LINEAR):
        """Changes an array of images to fit with model's input.
        Note that this method casts and rescales data to a range of [0, 1].
        
        Arguments:
            data_array {ndarray} -- An array of images.
        
        Keyword Arguments:
            cv_interpolation {int} -- cv2 interpolation. (default: {cv2.INTER_LINEAR})

        Returns:
            [ndarray] -- Reshaped data array.
        """
        if data_array.shape[1:] != self.model.input_shape:
            reshaped_list = list()
            for i in range(data_array.shape[0]):
                resized_img = cv2.resize(data_array[i], (self.model.input_shape[1], self.model.input_shape[2]), interpolation = cv_interpolation)
                reshaped_list.append(resized_img)
            data_array = np.asarray(reshaped_list)
            reshaped_list = [] # empty list after use
        data_array = data_array.astype('float32')
        data_array /= 255
        return data_array




"""Miscellaneous utility functions."""

from functools import reduce

from PIL import Image
import numpy as np
from matplotlib.colors import rgb_to_hsv, hsv_to_rgb

def compose(*funcs):
    """Compose arbitrarily many functions, evaluated left to right.

    Reference: https://mathieularose.com/function-composition-in-python/
    """
    # return lambda x: reduce(lambda v, f: f(v), funcs, x)
    if funcs:
        return reduce(lambda f, g: lambda *a, **kw: g(f(*a, **kw)), funcs)
    else:
        raise ValueError('Composition of empty sequence not supported.')

def letterbox_image(image, size):
    '''resize image with unchanged aspect ratio using padding'''
    iw, ih = image.size
    w, h = size
    scale = min(w/iw, h/ih)
    nw = int(iw*scale)
    nh = int(ih*scale)

    image = image.resize((nw,nh), Image.BICUBIC)
    new_image = Image.new('RGB', size, (128,128,128))
    new_image.paste(image, ((w-nw)//2, (h-nh)//2))
    return new_image

def rand(a=0, b=1):
    return np.random.rand()*(b-a) + a

def get_random_data(annotation_line, input_shape, random=True, max_boxes=20, jitter=.3, hue=.1, sat=1.5, val=1.5, proc_img=True):
    '''random preprocessing for real-time data augmentation'''
    line = annotation_line.split()  # split by SPACE. -libn
    image = Image.open(line[0])
    iw, ih = image.size
    h, w = input_shape
    box = np.array([np.array(list(map(int,box.split(',')))) for box in line[1:]])

    if not random:
        # resize image
        scale = min(w/iw, h/ih)
        nw = int(iw*scale)
        nh = int(ih*scale)
        dx = (w-nw)//2
        dy = (h-nh)//2
        image_data=0
        if proc_img:
            image = image.resize((nw,nh), Image.BICUBIC)
            new_image = Image.new('RGB', (w,h), (128,128,128))
            new_image.paste(image, (dx, dy))
            image_data = np.array(new_image)/255.

        # correct boxes
        box_data = np.zeros((max_boxes,5))
        if len(box)>0:
            np.random.shuffle(box)
            if len(box)>max_boxes: box = box[:max_boxes]
            box[:, [0,2]] = box[:, [0,2]]*scale + dx
            box[:, [1,3]] = box[:, [1,3]]*scale + dy
            box_data[:len(box)] = box

        return image_data, box_data

    # resize image
    new_ar = w/h * rand(1-jitter,1+jitter)/rand(1-jitter,1+jitter)
    scale = rand(.25, 2)
    if new_ar < 1:
        nh = int(scale*h)
        nw = int(nh*new_ar)
    else:
        nw = int(scale*w)
        nh = int(nw/new_ar)
    image = image.resize((nw,nh), Image.BICUBIC)

    # place image
    dx = int(rand(0, w-nw))
    dy = int(rand(0, h-nh))
    new_image = Image.new('RGB', (w,h), (128,128,128))
    new_image.paste(image, (dx, dy))
    image = new_image

    # flip image or not
    flip = rand()<.5
    if flip: image = image.transpose(Image.FLIP_LEFT_RIGHT)

    # distort image
    hue = rand(-hue, hue)
    sat = rand(1, sat) if rand()<.5 else 1/rand(1, sat)
    val = rand(1, val) if rand()<.5 else 1/rand(1, val)
    x = rgb_to_hsv(np.array(image)/255.)
    x[..., 0] += hue
    x[..., 0][x[..., 0]>1] -= 1
    x[..., 0][x[..., 0]<0] += 1
    x[..., 1] *= sat
    x[..., 2] *= val
    x[x>1] = 1
    x[x<0] = 0
    image_data = hsv_to_rgb(x) # numpy array, 0 to 1

    # correct boxes
    box_data = np.zeros((max_boxes,5))
    if len(box)>0:
        np.random.shuffle(box)
        box[:, [0,2]] = box[:, [0,2]]*nw/iw + dx
        box[:, [1,3]] = box[:, [1,3]]*nh/ih + dy
        if flip: box[:, [0,2]] = w - box[:, [2,0]]
        box[:, 0:2][box[:, 0:2]<0] = 0
        box[:, 2][box[:, 2]>w] = w
        box[:, 3][box[:, 3]>h] = h
        box_w = box[:, 2] - box[:, 0]
        box_h = box[:, 3] - box[:, 1]
        box = box[np.logical_and(box_w>1, box_h>1)] # discard invalid box
        if len(box)>max_boxes: box = box[:max_boxes]
        box_data[:len(box)] = box

    return image_data, box_data



# -*- coding: utf-8 -*-
"""
Class definition of YOLO_v3 style detection model on image and video
"""

import colorsys
import os
import cv2
from timeit import default_timer as timer
import requests
import numpy as np
from keras import backend as K
from keras.models import load_model, Model
from keras.layers import Input, Lambda
from keras.utils import multi_gpu_model
from keras.optimizers import Adam
from PIL import Image, ImageDraw
from ....lib.yolo_v3.model import yolo_eval, yolo_body, tiny_yolo_body, yolo_loss, preprocess_true_boxes
from ....lib.yolo_v3.utils import letterbox_image, get_random_data
from ...application import Application

class YOLO_V3(Application):
    def __init__(self, input_shape = None):
        if type(input_shape) != type(None):            
            self.input_shape = input_shape
        Application.__init__(self)
        self.anchors = np.array([10.0,13.0, 16.0,30.0, 33.0,23.0, 30.0,61.0, 62.0,45.0, 59.0,119.0, 116.0,90.0, 156.0,198.0, 373.0,326.0]).reshape(-1,2)
        self.sess = K.get_session()
        self.model_image_size = (416,416)

    def prepare_train_data(self, data_path = None, get_image_from = 'directory', data_array = None, classes_path = None, annotation_path = None, gpu_num = 1):
        self.data_path = data_path
        self.get_image_from = get_image_from
        self.data_array = data_array
        self.annotation_path = annotation_path
        self.gpu_num = gpu_num
        self.class_names = self.get_classes(classes_path)

        if get_image_from == 'directory':
            assert data_path, "Provide data_path to train from directory"
        elif get_image_from == 'argument':
            assert data_array, "Provide data_array to train from argument"
        else:
            raise ValueError("value of get_image_from should be either 'directory' or 'argument'.")

    def save(self, output_model_path):
        derived_model = Model(self.yolo_model.input[0], [self.yolo_model.layers[249].output, self.yolo_model.layers[250].output, self.yolo_model.layers[251].output])
        derived_model.save(output_model_path)
        print("saved model to " + output_model_path)

    def download_trained_weights(self, weights_path):
        print("Downloading pretrained model to " + weights_path + " ...")
        URL = "https://drive.google.com/uc?export=download"
        id = "1Dd-uUhhXvosXiIIZM8tiXoZyENJxIY4u"
        session = requests.Session()

        response = session.get(URL, params = { 'id' : id }, stream = True)
        token = None
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                token = value

        if token:
            params = { 'id' : id, 'confirm' : token }
            response = session.get(URL, params = params, stream = True)

        CHUNK_SIZE = 32768

        with open(weights_path, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)
        print("... done downloading pretrained model!")
        
    def create_model(self, input_shape, anchors, num_classes, load_pretrained=True, freeze_body=2, weights_path=None):
        '''create the training model'''
        K.clear_session() # get a new session
        image_input = Input(shape=(None, None, 3))
        h, w = input_shape
        num_anchors = len(anchors)

        # y_true = [Input(shape=(416//{0:32, 1:16, 2:8}[l], 416//{0:32, 1:16, 2:8}[l], 9//3, 80+5)) for l in range(3)]
        y_true = [Input(shape=(h//{0:32, 1:16, 2:8}[l], w//{0:32, 1:16, 2:8}[l], num_anchors//3, num_classes+5)) for l in range(3)]

        model_body = yolo_body(image_input, num_anchors//3, num_classes)
        print('Create YOLOv3 model with {} anchors and {} classes.'.format(num_anchors, num_classes))

        if load_pretrained:
            if not os.path.exists(weights_path):
                self.download_trained_weights(weights_path)
            model_body.load_weights(weights_path, by_name=True, skip_mismatch=True)
            print('Load weights {}.'.format(weights_path))
            if freeze_body in [1, 2]:
                # Freeze darknet53 body or freeze all but 3 output layers.
                num = (185, len(model_body.layers)-3)[freeze_body-1]
                for i in range(num): model_body.layers[i].trainable = False
                print('Freeze the first {} layers of total {} layers.'.format(num, len(model_body.layers)))

        model_loss = Lambda(yolo_loss, output_shape=(1,), name='yolo_loss',
            arguments={'anchors': anchors, 'num_classes': num_classes, 'ignore_thresh': 0.5})(
            [*model_body.output, *y_true])
        model = Model([model_body.input, *y_true], model_loss)
        print('model_body.input: ', model_body.input)
        print('model.input: ', model.input)

        return model

    def train(self, epochs = 50, batch_size = 32, weights_path = None):
        # return
        num_classes = len(self.class_names)
        anchors = self.anchors

        input_shape = (416,416) # multiple of 32, hw
        if hasattr(self, 'input_shape'):
            input_shape = self.input_shape

        is_tiny_version = len(anchors)==6 # default setting
        if is_tiny_version:
            model = self.create_tiny_model(input_shape, anchors, num_classes,
                freeze_body=2, weights_path=weights_path)
        else:
            model = self.create_model(input_shape, anchors, num_classes,
                freeze_body=2, weights_path=weights_path) # make sure you know what you freeze
        # model.save('yolo_model_retrain.h5')  # creates a HDF5 file 'my_model.h5'

        print(model.input)
        print(model.output)

        val_split = 0.1
        with open(self.annotation_path) as f:
            lines = f.readlines()
        np.random.seed(10101)
        np.random.shuffle(lines)
        np.random.seed(None)
        num_val = int(len(lines)*val_split)
        num_train = len(lines) - num_val


        # Train with frozen layers to get a stable loss.
        model.compile(optimizer=Adam(lr=1e-3), loss='mean_squared_error')

        print('Train on {} samples, val on {} samples, with batch size {}.'.format(num_train, num_val, batch_size))
        model.fit_generator(self.data_generator_wrapper(lines[:num_train], batch_size, input_shape, anchors, num_classes),
                steps_per_epoch=max(1, num_train//batch_size),
                validation_data=self.data_generator_wrapper(lines[num_train:], batch_size, input_shape, anchors, num_classes),
                validation_steps=max(1, num_val//batch_size),
                epochs=epochs,
                initial_epoch=0,
                callbacks=[])
        self.yolo_model = model

    def data_generator(self, annotation_lines, batch_size, input_shape, anchors, num_classes):
        '''data generator for fit_generator'''
        n = len(annotation_lines)
        i = 0
        while True:
            image_data = []
            box_data = []
            for b in range(batch_size):
                if i==0:
                    np.random.shuffle(annotation_lines)
                image, box = get_random_data(annotation_lines[i], input_shape, random=True)
                image_data.append(image)
                box_data.append(box)
                i = (i+1) % n
            image_data = np.array(image_data)   # input of original yolo: image
            box_data = np.array(box_data)       # output of original yolo: boxes
            y_true = preprocess_true_boxes(box_data, input_shape, anchors, num_classes) # some kind of output description?!
            yield [image_data, *y_true], np.zeros(batch_size)

    def data_generator_wrapper(self, annotation_lines, batch_size, input_shape, anchors, num_classes):
        n = len(annotation_lines)
        if n==0 or batch_size<=0:
            return None
        return self.data_generator(annotation_lines, batch_size, input_shape, anchors, num_classes)

    def create_tiny_model(self, input_shape, anchors, num_classes, load_pretrained=True, freeze_body=2,
                weights_path='model_data/tiny_yolo_weights.h5'):
        '''create the training model, for Tiny YOLOv3'''
        K.clear_session() # get a new session
        image_input = Input(shape=(None, None, 3))
        h, w = input_shape
        num_anchors = len(anchors)

        y_true = [Input(shape=(h//{0:32, 1:16}[l], w//{0:32, 1:16}[l], \
            num_anchors//2, num_classes+5)) for l in range(2)]

        model_body = tiny_yolo_body(image_input, num_anchors//2, num_classes)
        print('Create Tiny YOLOv3 model with {} anchors and {} classes.'.format(num_anchors, num_classes))

        if load_pretrained:
            model_body.load_weights(weights_path, by_name=True, skip_mismatch=True)
            print('Load weights {}.'.format(weights_path))
            if freeze_body in [1, 2]:
                # Freeze the darknet body or freeze all but 2 output layers.
                num = (20, len(model_body.layers)-2)[freeze_body-1]
                for i in range(num): model_body.layers[i].trainable = False
                print('Freeze the first {} layers of total {} layers.'.format(num, len(model_body.layers)))

        model_loss = Lambda(yolo_loss, output_shape=(1,), name='yolo_loss',
            arguments={'anchors': anchors, 'num_classes': num_classes, 'ignore_thresh': 0.7})(
            [*model_body.output, *y_true])
        model = Model([model_body.input, *y_true], model_loss)

        return model

    def get_classes(self, classes_path):
        '''loads the classes'''
        with open(classes_path) as f:
            class_names = f.readlines()
        class_names = [c.strip() for c in class_names]
        return class_names
        
    def generate(self, model_path = None, score_threshold = 0.3, iou_threshold = 0.45):
        model_path = os.path.expanduser(model_path)
        assert model_path.endswith('.h5'), 'Keras model or weights must be a .h5 file.'

        # Load model, or construct model and load weights.
        num_anchors = len(self.anchors)
        num_classes = len(self.class_names)
        is_tiny_version = num_anchors==6 # default setting
        try:
            self.yolo_model = load_model(model_path, compile=False)
        except:
            try:
                self.yolo_model = tiny_yolo_body(Input(shape=(None,None,3)), num_anchors//2, num_classes) \
                    if is_tiny_version else yolo_body(Input(shape=(None,None,3)), num_anchors//3, num_classes)
                self.yolo_model.load_weights(model_path) # make sure model, anchors and classes match
            except:
                if not os.path.exists(model_path):
                    self.download_trained_weights(model_path)
                else:
                    print("Cannot load the model")
        else:
            print('output_shape = %d' %(self.yolo_model.layers[-1].output_shape[-1]))
            print('num_anchors = %d' % num_anchors)
            print('len = %d' %(len(self.yolo_model.output) * (num_classes + 5)))
            print('len_output = %d' %(len(self.yolo_model.output)))

        print('{} model, anchors, and classes loaded.'.format(model_path))

        # Generate colors for drawing bounding boxes.
        hsv_tuples = [(x / len(self.class_names), 1., 1.)
                      for x in range(len(self.class_names))]
        self.colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        self.colors = list(
            map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
                self.colors))
        np.random.seed(10101)  # Fixed seed for consistent colors across runs.
        np.random.shuffle(self.colors)  # Shuffle colors to decorrelate adjacent classes.
        np.random.seed(None)  # Reset seed to default.

        # Generate output tensor targets for filtered bounding boxes.
        self.input_image_shape = K.placeholder(shape=(2, ))
        if self.gpu_num>=2:
            self.yolo_model = multi_gpu_model(self.yolo_model, gpus=self.gpu_num)
        self.boxes, self.scores, self.classes = yolo_eval(self.yolo_model.output, self.anchors,
                len(self.class_names), self.input_image_shape,
                score_threshold=score_threshold, iou_threshold=iou_threshold)

    def detect_image(self, image):
        image = Image.open(image)
        image = self._detect_image(image)
        return image

    def _detect_image(self, image):
        start = timer()

        if self.model_image_size != (None, None):
            assert self.model_image_size[0]%32 == 0, 'Multiples of 32 required'
            assert self.model_image_size[1]%32 == 0, 'Multiples of 32 required'
            boxed_image = letterbox_image(image, tuple(reversed(self.model_image_size)))
        else:
            new_image_size = (image.width - (image.width % 32),
                              image.height - (image.height % 32))
            boxed_image = letterbox_image(image, new_image_size)
        image_data = np.array(boxed_image, dtype='float32')

        print(image_data.shape)
        image_data /= 255.
        image_data = np.expand_dims(image_data, 0)  # Add batch dimension.

        out_boxes, out_scores, out_classes = self.sess.run(
            [self.boxes, self.scores, self.classes],
            feed_dict={
                self.yolo_model.input: image_data,
                self.input_image_shape: [image.size[1], image.size[0]],
                K.learning_phase(): 0
            })

        print('Found {} boxes for {}'.format(len(out_boxes), 'img'))

        thickness = (image.size[0] + image.size[1]) // 300

        for i, c in reversed(list(enumerate(out_classes))):
            predicted_class = self.class_names[c]
            box = out_boxes[i]
            score = out_scores[i]

            label = '{} {:.2f}'.format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label)

            top, left, bottom, right = box
            top = max(0, np.floor(top + 0.5).astype('int32'))
            left = max(0, np.floor(left + 0.5).astype('int32'))
            bottom = min(image.size[1], np.floor(bottom + 0.5).astype('int32'))
            right = min(image.size[0], np.floor(right + 0.5).astype('int32'))
            print(label, (left, top), (right, bottom))

            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, top + 1])

            # My kingdom for a good redistributable image drawing library.
            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i],
                    outline=self.colors[c])
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=self.colors[c])
            print(text_origin.shape)
            draw.text(list(text_origin), label, fill=(0, 0, 0))
            del draw

        end = timer()
        print(end - start)
        image.show()

        return image
    
    def save_image(self, image, output_path):
        opencvImage = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
        cv2.imwrite(output_path, opencvImage)

    def detect_video(self, video_path, output_path=""):
        vid = cv2.VideoCapture(video_path)
        print('vid: ', vid)
        print('output_path: ', output_path)
        if not vid.isOpened():
            raise IOError("Couldn't open webcam or video")
        video_FourCC    = cv2.VideoWriter_fourcc(*"mp4v")
        video_fps       = vid.get(cv2.CAP_PROP_FPS)
        video_size      = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)),
                            int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        isOutput = True if output_path != "" else False
        if isOutput:
            print("!!! TYPE:", type(output_path), type(video_FourCC), type(video_fps), type(video_size))
            out = cv2.VideoWriter(output_path, video_FourCC, video_fps, video_size)
        accum_time = 0
        curr_fps = 0
        fps = "FPS: ??"
        prev_time = timer()
        while True:
            return_value, frame = vid.read()
            if not return_value:
                break
            image = Image.fromarray(frame)
            image = self._detect_image(image)
            result = np.asarray(image)
            curr_time = timer()
            exec_time = curr_time - prev_time
            prev_time = curr_time
            accum_time = accum_time + exec_time
            curr_fps = curr_fps + 1
            if accum_time > 1:
                accum_time = accum_time - 1
                fps = "FPS: " + str(curr_fps)
                curr_fps = 0
            cv2.putText(result, text=fps, org=(3, 15), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=0.50, color=(255, 0, 0), thickness=2)
            cv2.namedWindow("result", cv2.WINDOW_NORMAL)
            cv2.imshow("result", result)
            if isOutput:
                out.write(result)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        self.sess.close()


from keras.models import load_model

import numpy
import cv2

import os
import platform

class Application(object):
    
    def __init__(self):
        try:
            self.os_type = platform.system()
        except Exception as e:
            print("Application initialize failed")
            raise e
        # default setting for using 1gpu and gpu_ids for 0
        self.gpu_num = 1
        self.gpu_ids = [0]

    def use_specific_gpus(self, gpuids):
        self.gpu_ids = gpuids
        self.gpu_num = len(gpuids)
        



import keras
import numpy as np
import os
import cv2
"""This module is a custom module designed for general use in the Omnis project.
"""

class DataGenerator(keras.utils.Sequence):
    def __init__(self, image_filenames = None, labels=None, batch_size=None, image_shape=None, shuffle = False):
        self.image_filenames, self.labels = image_filenames, labels
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle
        self.predict_order_filenames = []

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[idx * self.batch_size:(idx + 1) * self.batch_size]

        x = []
        for filename in batch_x:
            _, only_file_name = os.path.split(filename)
            self.predict_order_filenames.append(only_file_name)
            img = cv2.imread(filename)
            test_data = np.expand_dims(img, axis=0)
            resized_img = cv2.resize(test_data[0], self.image_shape)
            x.append(resized_img)
        
        data_array = np.asarray(x)
        data_array = data_array.astype('float32')
        data_array /= 255
        return data_array


    def flow_one_directory(self, directory_path, shuffle = False, batch_size = 1, image_shape = None):
        filelist = os.listdir(directory_path)
        for i in range(len(filelist)):
            if i==0:
                self.image_filenames = [directory_path + '/' + filelist[i]]
            else :
                self.image_filenames.append(directory_path + '/' + filelist[i])
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle
        
        return self


def reverse_dict(input_dict):
    """This function returns a dictionary with input_dict's key, value is reversed.
    
    Arguments:
        input_dict {dict} -- Dictionary to reverse.
    
    Returns:
        [dict] -- Reversed dictionary.
    """
    reversed_dict = {}
    key_list = list(input_dict.keys())
    for index, val in enumerate( list(input_dict.values()) ):
        reversed_dict[val] = key_list[index]
    return reversed_dict


from keras import backend as K
import keras

from keras.preprocessing.image import load_img, img_to_array
import numpy as np
from scipy.optimize import fmin_l_bfgs_b
import time
import os
import skimage
import cv2
from ...application import Application

from multiprocessing import Process, Queue
from .style_transfer_worker import Style_Transfer_Worker

class Style_Transfer(Application):
    def __init__(self):
        super().__init__()
        
        # This application does not support Window now
        if self.os_type == "Window":
            raise ValueError("This application does not support window yet")
        self.loss_value = None
        self.grad_values = None
        self._queue = Queue()


    def prepare_train_data(self, get_content_image_from = 'directory', content_image_path = None, get_style_image_as = 'filepath', style_image_path = None):
        assert get_content_image_from in ["directory", "argument"], "get_content_image_from should be either 'directory' or 'argument'."
        assert get_style_image_as in ["filepath", "ndarray"], "get_style_image_as should be either 'filepath' or 'ndarray'."

        self._workers = list()
        for gpuid in self.gpu_ids:
            self._workers.append(Style_Transfer_Worker(
                gpuid = gpuid, 
                queue = self._queue,
                get_content_image_from = get_content_image_from,
                get_style_image_as = get_style_image_as,
                content_image_path = content_image_path,
                style_image_path = style_image_path))

        # scan all files under get_content_image_path
        for content_file in os.listdir(content_image_path):
            self._queue.put(os.path.join(content_image_path, content_file))
        self._queue.put(None)

    def generate(self, iterations = 10, output_type = "file", output_path = "result", total_variation_weight=1.0, style_weight=1.0, content_weight=0.025):
        for worker in self._workers:
            worker.output_type = output_type
            worker.output_path = output_path
            worker.iterations = iterations
            worker.style_weight = style_weight
            worker.content_weight = content_weight
            worker.total_variation_weight = total_variation_weight

        for worker in self._workers:
            worker.start()
        
        for worker in self._workers:
            worker.join()
        
        print("all of workers have been done")
